"""Data Handler using kz-filtered data."""

__author__ = 'Lukas Leufen'
__date__ = '2020-08-26'

import copy
import numpy as np
import pandas as pd
import xarray as xr
from typing import List, Union, Tuple, Optional
from functools import partial
import logging
from mlair.data_handler.data_handler_single_station import DataHandlerSingleStation
from mlair.data_handler import DefaultDataHandler
from mlair.helpers import to_list, TimeTrackingWrapper, statistics
from mlair.helpers.filter import FIRFilter, ClimateFIRFilter, omega_null_kzf

# define a more general date type for type hinting
str_or_list = Union[str, List[str]]


# cutoff_p = [(None, 14), (8, 6), (2, 0.8), (0.8, None)]
# cutoff = list(map(lambda x: (1. / x[0] if x[0] is not None else None, 1. / x[1] if x[1] is not None else None), cutoff_p))
# fs = 24.
# # order = int(60 * fs) + 1
# order = np.array([int(14 * fs) + 1, int(14 * fs) + 1, int(4 * fs) + 1, int(2 * fs) + 1])
# print("cutoff period", cutoff_p)
# print("cutoff", cutoff)
# print("fs", fs)
# print("order", order)
# print("delay", 0.5 * (order-1) / fs)
# window = ("kaiser", 5)
# # low pass
# y, h = fir_filter(station_data.values.flatten(), fs, order[0], cutoff_low = cutoff[0][0], cutoff_high = cutoff[0][1], window=window)
# filtered = xr.ones_like(station_data) * y.reshape(station_data.values.shape)


class DataHandlerFilterSingleStation(DataHandlerSingleStation):
    """General data handler for a single station to be used by a superior data handler."""

    _hash = DataHandlerSingleStation._hash + ["filter_dim"]

    DEFAULT_FILTER_DIM = "filter"

    def __init__(self, *args, filter_dim=DEFAULT_FILTER_DIM, **kwargs):
        # self.original_data = None  # ToDo: implement here something to store unfiltered data
        self.filter_dim = filter_dim
        self.filter_dim_order = None
        super().__init__(*args, **kwargs)

    def setup_transformation(self, transformation: Union[None, dict, Tuple]) -> Tuple[Optional[dict], Optional[dict]]:
        """
        Adjust setup of transformation because filtered data will have negative values which is not compatible with
        the log transformation. Therefore, replace all log transformation methods by a default standardization. This is
        only applied on input side.
        """
        transformation = super(__class__, self).setup_transformation(transformation)
        if transformation[0] is not None:
            for k, v in transformation[0].items():
                if v["method"] == "log":
                    transformation[0][k]["method"] = "standardise"
                elif v["method"] == "min_max":
                    transformation[0][k]["method"] = "standardise"
        return transformation

    def _check_sampling(self, **kwargs):
        assert kwargs.get("sampling") == "hourly"  # This data handler requires hourly data resolution, does it?

    def make_input_target(self):
        data, self.meta = self.load_data(self.path, self.station, self.statistics_per_var, self.sampling,
                                         self.store_data_locally, self.data_origin, self.start, self.end)
        self._data = self.interpolate(data, dim=self.time_dim, method=self.interpolation_method,
                                      limit=self.interpolation_limit)
        self.set_inputs_and_targets()
        self.apply_filter()
        # this is just a code snippet to check the results of the kz filter
        # import matplotlib
        # matplotlib.use("TkAgg")
        # import matplotlib.pyplot as plt
        # self.input_data.sel(filter="low", variables="temp", Stations="DEBW107").plot()
        # self.input_data.sel(variables="temp", Stations="DEBW107").plot.line(hue="filter")

    def apply_filter(self):
        raise NotImplementedError

    def create_filter_index(self) -> pd.Index:
        """Create name for filter dimension."""
        raise NotImplementedError

    def get_transposed_history(self) -> xr.DataArray:
        """Return history.

        :return: history with dimensions datetime, window, Stations, variables, filter.
        """
        return self.history.transpose(self.time_dim, self.window_dim, self.iter_dim, self.target_dim,
                                      self.filter_dim).copy()

    def _create_lazy_data(self):
        raise NotImplementedError

    def _extract_lazy(self, lazy_data):
        _data, self.meta, _input_data, _target_data = lazy_data
        f_prep = partial(self._slice_prep, start=self.start, end=self.end)
        self._data, self.input_data, self.target_data = list(map(f_prep, [_data, _input_data, _target_data]))


class DataHandlerFilter(DefaultDataHandler):
    """Data handler using FIR filtered data."""

    data_handler = DataHandlerFilterSingleStation
    data_handler_transformation = DataHandlerFilterSingleStation
    _requirements = data_handler.requirements()

    def __init__(self, *args, use_filter_branches=False, **kwargs):
        self.use_filter_branches = use_filter_branches
        super().__init__(*args, **kwargs)

    def get_X_original(self):
        if self.use_filter_branches is True:
            X = []
            for data in self._collection:
                if hasattr(data, "filter_dim"):
                    X_total = data.get_X()
                    filter_dim = data.filter_dim
                    for filter_name in data.filter_dim_order:
                        X.append(X_total.sel({filter_dim: filter_name}, drop=True))
                else:
                    X.append(data.get_X())
            return X
        else:
            return super().get_X_original()


class DataHandlerFirFilterSingleStation(DataHandlerFilterSingleStation):
    """Data handler for a single station to be used by a superior data handler. Inputs are FIR filtered."""

    _hash = DataHandlerFilterSingleStation._hash + ["filter_cutoff_period", "filter_order", "filter_window_type"]

    DEFAULT_WINDOW_TYPE = ("kaiser", 5)

    def __init__(self, *args, filter_cutoff_period, filter_order, filter_window_type=DEFAULT_WINDOW_TYPE,
                 plot_path=None, filter_plot_dates=None, **kwargs):
        # self.original_data = None  # ToDo: implement here something to store unfiltered data
        self.fs = self._get_fs(**kwargs)
        if filter_window_type == "kzf":
            filter_cutoff_period = self._get_kzf_cutoff_period(filter_order, self.fs)
        self.filter_cutoff_period, removed_index = self._prepare_filter_cutoff_period(filter_cutoff_period, self.fs)
        self.filter_cutoff_freq = self._period_to_freq(self.filter_cutoff_period)
        assert len(self.filter_cutoff_period) == (len(filter_order) - len(removed_index))
        self.filter_order = self._prepare_filter_order(filter_order, removed_index, self.fs)
        self.filter_window_type = filter_window_type
        self.unfiltered_name = "unfiltered"
        self.plot_path = plot_path  # use this path to create insight plots
        self.plot_dates = filter_plot_dates
        super().__init__(*args, **kwargs)

    @staticmethod
    def _prepare_filter_order(filter_order, removed_index, fs):
        order = []
        for i, o in enumerate(filter_order):
            if i not in removed_index:
                if isinstance(o, tuple):
                    fo = (o[0] * fs, o[1])
                else:
                    fo = int(o * fs)
                    fo = fo + 1 if fo % 2 == 0 else fo
                order.append(fo)
        return order

    @staticmethod
    def _prepare_filter_cutoff_period(filter_cutoff_period, fs):
        """Frequency must be smaller than the sampling frequency fs. Otherwise remove given cutoff period pair."""
        cutoff = []
        removed = []
        for i, period in enumerate(to_list(filter_cutoff_period)):
            if period > 2. / fs:
                cutoff.append(period)
            else:
                removed.append(i)
        return cutoff, removed

    @staticmethod
    def _get_kzf_cutoff_period(kzf_settings, fs):
        cutoff = []
        for (m, k) in kzf_settings:
            w0 = omega_null_kzf(m * fs, k) * fs
            cutoff.append(1. / w0)
        return cutoff

    @staticmethod
    def _period_to_freq(cutoff_p):
        return [1. / x for x in cutoff_p]

    @staticmethod
    def _get_fs(**kwargs):
        """Return frequency in 1/day (not Hz)"""
        sampling = kwargs.get("sampling")
        if sampling == "daily":
            return 1
        elif sampling == "hourly":
            return 24
        else:
            raise ValueError(f"Unknown sampling rate {sampling}. Only daily and hourly resolution is supported.")

    @TimeTrackingWrapper
    def apply_filter(self):
        """Apply FIR filter only on inputs."""
        fir = FIRFilter(self.input_data.astype("float32"), self.fs, self.filter_order, self.filter_cutoff_freq,
                        self.filter_window_type, self.target_dim, self.time_dim, display_name=self.station[0],
                        minimum_length=self.window_history_size, offset=self.window_history_offset,
                        plot_path=self.plot_path, plot_dates=self.plot_dates)
        self.fir_coeff = fir.filter_coefficients
        filter_data = fir.filtered_data
        input_data = xr.concat(filter_data, pd.Index(self.create_filter_index(), name=self.filter_dim))
        self.input_data = input_data.sel({self.target_dim: self.variables})
        # this is just a code snippet to check the results of the kz filter
        # import matplotlib
        # matplotlib.use("TkAgg")
        # import matplotlib.pyplot as plt
        # self.input_data.sel(filter="low", variables="temp", Stations="DEBW107").plot()
        # self.input_data.sel(variables="temp", Stations="DEBW107").plot.line(hue="filter")

    def create_filter_index(self, add_unfiltered_index=True) -> pd.Index:
        """
        Round cut off periods in days and append 'res' for residuum index.

        Round small numbers (<10) to single decimal, and higher numbers to int. Transform as list of str and append
        'res' for residuum index. Add index unfiltered if the raw / unfiltered data is appended to data in addition.
        """
        index = np.round(self.filter_cutoff_period, 1)
        f = lambda x: int(np.round(x)) if x >= 10 else np.round(x, 1)
        index = list(map(f, index.tolist()))
        index = list(map(lambda x: str(x) + "d", index)) + ["res"]
        self.filter_dim_order = index
        return pd.Index(index, name=self.filter_dim)

    def _create_lazy_data(self):
        return [self._data, self.meta, self.input_data, self.target_data, self.fir_coeff, self.filter_dim_order]

    def _extract_lazy(self, lazy_data):
        _data, _meta, _input_data, _target_data, self.fir_coeff, self.filter_dim_order = lazy_data
        super()._extract_lazy((_data, _meta, _input_data, _target_data))

    def transform(self, data_in, dim: Union[str, int] = 0, inverse: bool = False, opts=None,
                  transformation_dim=None):
        """
        Transform data according to given transformation settings.

        This function transforms a xarray.dataarray (along dim) or pandas.DataFrame (along axis) either with mean=0
        and std=1 (`method=standardise`) or centers the data with mean=0 and no change in data scale
        (`method=centre`). Furthermore, this sets an internal instance attribute for later inverse transformation. This
        method will raise an AssertionError if an internal transform method was already set ('inverse=False') or if the
        internal transform method, internal mean and internal standard deviation weren't set ('inverse=True').

        :param string/int dim: This param is not used for inverse transformation.
                | for xarray.DataArray as string: name of dimension which should be standardised
                | for pandas.DataFrame as int: axis of dimension which should be standardised
        :param inverse: Switch between transformation and inverse transformation.

        :return: xarray.DataArrays or pandas.DataFrames:
                #. mean: Mean of data
                #. std: Standard deviation of data
                #. data: Standardised data
        """

        if transformation_dim is None:
            transformation_dim = self.DEFAULT_TARGET_DIM

        def f(data, method="standardise", feature_range=None):
            if method == "standardise":
                return statistics.standardise(data, dim)
            elif method == "centre":
                return statistics.centre(data, dim)
            elif method == "min_max":
                kwargs = {"feature_range": feature_range} if feature_range is not None else {}
                return statistics.min_max(data, dim, **kwargs)
            elif method == "log":
                return statistics.log(data, dim)
            else:
                raise NotImplementedError

        def f_apply(data, method, **kwargs):
            for k, v in kwargs.items():
                if not (isinstance(v, xr.DataArray) or v is None):
                    _, opts = statistics.min_max(data, dim)
                    helper = xr.ones_like(opts['min'])
                    kwargs[k] = helper * v
            mean = kwargs.pop('mean', None)
            std = kwargs.pop('std', None)
            min = kwargs.pop('min', None)
            max = kwargs.pop('max', None)
            feature_range = kwargs.pop('feature_range', None)

            if method == "standardise":
                return statistics.standardise_apply(data, mean, std), {"mean": mean, "std": std, "method": method}
            elif method == "centre":
                return statistics.centre_apply(data, mean), {"mean": mean, "method": method}
            elif method == "min_max":
                return statistics.min_max_apply(data, min, max), {"min": min, "max": max, "method": method,
                                                                  "feature_range": feature_range}
            elif method == "log":
                return statistics.log_apply(data, mean, std), {"mean": mean, "std": std, "method": method}
            else:
                raise NotImplementedError

        opts = opts or {}
        opts_updated = {}
        if not inverse:
            transformed_values = []
            for var in data_in.variables.values:
                data_var = data_in.sel(**{transformation_dim: [var]})
                var_opts = opts.get(var, {})
                _apply = (var_opts.get("mean", None) is not None) or (var_opts.get("min") is not None)
                values, new_var_opts = locals()["f_apply" if _apply else "f"](data_var, **var_opts)
                opts_updated[var] = copy.deepcopy(new_var_opts)
                transformed_values.append(values)
            return xr.concat(transformed_values, dim=transformation_dim), opts_updated
        else:
            return self.inverse_transform(data_in, opts, transformation_dim)


class DataHandlerFirFilter(DataHandlerFilter):
    """Data handler using FIR filtered data."""

    data_handler = DataHandlerFirFilterSingleStation
    data_handler_transformation = DataHandlerFirFilterSingleStation
    _requirements = data_handler.requirements()


class DataHandlerClimateFirFilterSingleStation(DataHandlerFirFilterSingleStation):
    """
    Data handler for a single station to be used by a superior data handler. Inputs are FIR filtered. In contrast to
    the simple DataHandlerFirFilterSingleStation, this data handler is centered around t0 to have no time delay. For
    values in the future (t > t0), this data handler assumes a climatological value for the low pass data and values of
    0 for all residuum components.

    :param apriori: Data to use as apriori information. This should be either a xarray dataarray containing monthly or
        any other heuristic to support the clim filter, or a list of such arrays containing heuristics for all residua
        in addition. The 2nd can be used together with apriori_type `residuum_stats` which estimates the error of the
        residuum when the clim filter should be applied with exogenous parameters. If apriori_type is None/`zeros` data
        can be provided, but this is not required in this case.
    :param apriori_type: set type of information that is provided to the clim filter. For the first low pass always a
        calculated or given statistic is used. For residuum prediction a constant value of zero is assumed if
        apriori_type is None or `zeros`, and a climatology of the residuum is used for `residuum_stats`.
    :param apriori_diurnal: use diurnal anomalies of each hour as addition to the apriori information type chosen by
        parameter apriori_type. This is only applicable for hourly resolution data.
    :param apriori_sel_opts: specify some parameters to select a subset of data before calculating the apriori
        information. Use this parameter for example, if apriori shall only calculated on a shorter time period than
        available in given data.
    :param extend_length_opts: use this parameter to use future data in the filter calculation. This parameter does not
        affect the size of the history samples as this is handled by the window_history_size parameter. Example: set
        extend_length_opts=7*24 to use the observation of the next 7 days to calculate the filtered components. Which
        data are finally used for the input samples is not affected by these 7 days. In case the range of history sample
        exceeds the horizon of extend_length_opts, the history sample will also include data from climatological
        estimates.
    """
    DEFAULT_EXTEND_LENGTH_OPTS = 0
    _hash = DataHandlerFirFilterSingleStation._hash + ["apriori_type", "apriori_sel_opts", "apriori_diurnal",
                                                       "extend_length_opts"]
    _store_attributes = DataHandlerFirFilterSingleStation.store_attributes() + ["apriori"]

    def __init__(self, *args, apriori=None, apriori_type=None, apriori_diurnal=False, apriori_sel_opts=None,
                 extend_length_opts=DEFAULT_EXTEND_LENGTH_OPTS, **kwargs):
        self.apriori_type = apriori_type
        self.climate_filter_coeff = None  # coefficents of the used FIR filter
        self.apriori = apriori  # exogenous apriori information or None to calculate from data (endogenous)
        self.apriori_diurnal = apriori_diurnal
        self.all_apriori = None  # collection of all apriori information
        self.apriori_sel_opts = apriori_sel_opts  # ensure to separate exogenous and endogenous information
        self.extend_length_opts = extend_length_opts
        super().__init__(*args, **kwargs)

    @TimeTrackingWrapper
    def apply_filter(self):
        """Apply FIR filter only on inputs."""
        self.apriori = self.apriori.get(str(self)) if isinstance(self.apriori, dict) else self.apriori
        logging.info(f"{self.station}: call ClimateFIRFilter")
        climate_filter = ClimateFIRFilter(self.input_data.astype("float32"), self.fs, self.filter_order,
                                          self.filter_cutoff_freq,
                                          self.filter_window_type, time_dim=self.time_dim, var_dim=self.target_dim,
                                          apriori_type=self.apriori_type, apriori=self.apriori,
                                          apriori_diurnal=self.apriori_diurnal, sel_opts=self.apriori_sel_opts,
                                          plot_path=self.plot_path,
                                          minimum_length=self.window_history_size, new_dim=self.window_dim,
                                          display_name=self.station[0], extend_length_opts=self.extend_length_opts,
                                          extend_end=self.window_history_end, plot_dates=self.plot_dates,
                                          offset=self.window_history_offset)
        self.climate_filter_coeff = climate_filter.filter_coefficients

        # store apriori information: store all if residuum_stat method was used, otherwise just store initial apriori
        if self.apriori_type == "residuum_stats":
            self.apriori = climate_filter.apriori_data
        else:
            self.apriori = climate_filter.initial_apriori_data
        self.all_apriori = climate_filter.apriori_data

        climate_filter_data = [c.sel({self.window_dim: slice(self.window_history_end-self.window_history_size,
                                                             self.window_history_end)})
                               for c in climate_filter.filtered_data]

        # create input data with filter index
        input_data = xr.concat(climate_filter_data, pd.Index(self.create_filter_index(add_unfiltered_index=False),
                                                             name=self.filter_dim))

        self.input_data = input_data.sel({self.target_dim: self.variables})

        # this is just a code snippet to check the results of the filter
        # import matplotlib
        # matplotlib.use("TkAgg")
        # import matplotlib.pyplot as plt
        # self.input_data.sel(filter="low", variables="temp", Stations="DEBW107").plot()
        # self.input_data.sel(variables="temp", Stations="DEBW107").plot.line(hue="filter")

    def create_filter_index(self, add_unfiltered_index=True) -> pd.Index:
        """
        Round cut off periods in days and append 'res' for residuum index.

        Round small numbers (<10) to single decimal, and higher numbers to int. Transform as list of str and append
        'res' for residuum index. Add index unfiltered if the raw / unfiltered data is appended to data in addition.
        """
        index = np.round(self.filter_cutoff_period, 1)
        f = lambda x: int(np.round(x)) if x >= 10 else np.round(x, 1)
        index = list(map(f, index.tolist()))
        index = list(map(lambda x: str(x) + "d", index)) + ["res"]
        self.filter_dim_order = index
        return pd.Index(index, name=self.filter_dim)

    def _create_lazy_data(self):
        return [self._data, self.meta, self.input_data, self.target_data, self.climate_filter_coeff,
                self.apriori, self.all_apriori, self.filter_dim_order]

    def _extract_lazy(self, lazy_data):
        _data, _meta, _input_data, _target_data, self.climate_filter_coeff, self.apriori, self.all_apriori, \
        self.filter_dim_order = lazy_data
        DataHandlerSingleStation._extract_lazy(self, (_data, _meta, _input_data, _target_data))

    @staticmethod
    def _prepare_filter_cutoff_period(filter_cutoff_period, fs):
        """Frequency must be smaller than the sampling frequency fs. Otherwise remove given cutoff period pair."""
        cutoff = []
        removed = []
        for i, period in enumerate(to_list(filter_cutoff_period)):
            if period > 2. / fs:
                cutoff.append(period)
            else:
                removed.append(i)
        return cutoff, removed

    @staticmethod
    def _period_to_freq(cutoff_p):
        return [1. / x for x in cutoff_p]

    def make_history_window(self, dim_name_of_inputs: str, window: int, dim_name_of_shift: str) -> None:
        """
        Create a xr.DataArray containing history data. As 'input_data' already consists of a dimension 'window', this
        method only shifts the data along 'window' dimension x times where x is given by 'window_history_offset'.
        Results are stored in history attribute.

        :param dim_name_of_inputs: Name of dimension which contains the input variables
        :param window: this parameter is not used in the inherited method
        :param dim_name_of_shift: Dimension along shift will be applied
        """
        self.history = self.input_data
        # from matplotlib import pyplot as plt
        # d = self.load_and_interpolate(0)
        # data.sel(datetime="2007-07-07 00:00").sum("filter").plot()
        # plt.plot(data.sel(datetime="2007-07-07 00:00").sum("filter").window.values, d.sel(datetime=slice("2007-07-05 00:00", "2007-07-07 16:00")).values.flatten())
        # plt.plot(data.sel(datetime="2007-07-07 00:00").sum("filter").window.values, d.sel(datetime=slice("2007-07-05 00:00", "2007-07-11 16:00")).values.flatten())

    def call_transform(self, inverse=False):
        opts_input = self._transformation[0]
        self.input_data, opts_input = self.transform(self.input_data, dim=[self.time_dim, self.window_dim],
                                                     inverse=inverse, opts=opts_input,
                                                     transformation_dim=self.target_dim)
        opts_target = self._transformation[1]
        self.target_data, opts_target = self.transform(self.target_data, dim=self.time_dim, inverse=inverse,
                                                       opts=opts_target, transformation_dim=self.target_dim)
        self._transformation = (opts_input, opts_target)


class DataHandlerClimateFirFilter(DataHandlerFilter):
    """Data handler using climatic adjusted FIR filtered data."""

    data_handler = DataHandlerClimateFirFilterSingleStation
    data_handler_transformation = DataHandlerClimateFirFilterSingleStation
    _requirements = data_handler.requirements()
    _store_attributes = data_handler.store_attributes()

