"""Helper functions that are used to simplify testing."""
import logging
import re
from typing import Union, Pattern, List
import inspect

import numpy as np
import xarray as xr

from mlair.helpers.helpers import remove_items, to_list


class PyTestRegex:
    r"""
    Assert that a given string meets some expectations.

    Use like

        >>> PyTestRegex(r"TestString\d+") == "TestString"
        False
        >>> PyTestRegex(r"TestString\d+") == "TestString2"
        True


    :param pattern: pattern or string to use for regular expresssion
    :param flags: python re flags
    """

    def __init__(self, pattern: Union[str, Pattern], flags: int = 0):
        """Construct PyTestRegex."""
        self._regex = re.compile(pattern, flags)

    def __eq__(self, actual: str) -> bool:
        """Return whether regex matches given string actual or not."""
        return bool(self._regex.match(actual))

    def __repr__(self) -> str:
        """Show regex pattern."""
        return self._regex.pattern


def PyTestAllEqual(check_list: List):
    class PyTestAllEqualClass:
        """
        Check if all elements in list are the same.

        :param check_list: list with elements to check
        """

        def __init__(self, check_list: List):
            """Construct class."""
            self._list = check_list
            self._test_function = None

        def _set_test_function(self, _list):
            if isinstance(_list[0], list):
                _test_function = self._set_test_function(_list[0])
                self._test_function = lambda r, s: all(map(lambda x, y: _test_function(x, y) is None, r, s))
            elif isinstance(_list[0], np.ndarray):
                self._test_function = np.testing.assert_array_equal
            elif isinstance(_list[0], xr.DataArray):
                self._test_function = xr.testing.assert_equal
            else:
                self._test_function = lambda x, y: self._assert(x, y)
                # raise TypeError(f"given type {type(_list[0])} is not supported by PyTestAllEqual.")
            return self._test_function

        @staticmethod
        def _assert(x, y):
            assert x == y

        def _check_all_equal(self) -> bool:
            """
            Check if all elements are equal.

            :return boolean if elements are equal
            """
            equal = True
            self._set_test_function(self._list)
            for b in self._list:
                equal *= self._test_function(self._list[0], b) in [None, True]
            return bool(equal == 1)

        def is_true(self) -> bool:
            """
            Start equality check.

            :return: true if equality test is passed, false otherwise
            """
            return self._check_all_equal()

    return PyTestAllEqualClass(check_list).is_true()


def get_all_args(*args, remove=None, add=None):
    res = []
    for a in args:
        arg_spec = inspect.getfullargspec(a)
        res.extend(arg_spec.args)
        res.extend(arg_spec.kwonlyargs)
    res = sorted(list(set(res)))
    if remove is not None:
        res = remove_items(res, remove)
    if add is not None:
        res += to_list(add)
    return res


def check_nested_equality(obj1, obj2, precision=None, skip_args=None):
    """Check for equality in nested structures. Use precision to indicate number of decimals to check for consistency"""

    assert precision is None or isinstance(precision, int)
    message = ""
    try:
        # print(f"check type {type(obj1)} and {type(obj2)}")
        message = f"{type(obj1)}!={type(obj2)}\n{obj1} and {obj2} do not match"
        assert type(obj1) == type(obj2)
        if isinstance(obj1, (tuple, list)):
            # print(f"check length {len(obj1)} and {len(obj2)}")
            message = f"{len(obj1)}!={len(obj2)}\nlengths of {obj1} and {obj2} do not match"
            assert len(obj1) == len(obj2)
            for pos in range(len(obj1)):
                # print(f"check pos {obj1[pos]} and {obj2[pos]}")
                message = f"{obj1[pos]}!={obj2[pos]}\nobjects on pos {pos} of {obj1} and {obj2} do not match"
                assert check_nested_equality(obj1[pos], obj2[pos], precision=precision, skip_args=skip_args) is True
        elif isinstance(obj1, dict):
            obj1_keys, obj2_keys = obj1.keys(), obj2.keys()
            if skip_args is not None and isinstance(skip_args, (str, list)):
                skip_args = to_list(skip_args)
                obj1_keys = list(set(obj1_keys).difference(skip_args))
                obj2_keys = list(set(obj2_keys).difference(skip_args))
            # print(f"check keys {obj1.keys()} and {obj2.keys()}")
            message = f"{sorted(obj1_keys)}!={sorted(obj2_keys)}\n{set(obj1_keys).symmetric_difference(obj2_keys)} " \
                      f"are not in both sorted key lists"
            assert sorted(obj1_keys) == sorted(obj2_keys)
            for k in obj1_keys:
                # print(f"check pos {obj1[k]} and {obj2[k]}")
                message = f"{obj1[k]}!={obj2[k]}\nobjects for key {k} of {obj1} and {obj2} do not match"
                assert check_nested_equality(obj1[k], obj2[k], precision=precision, skip_args=skip_args) is True
        elif isinstance(obj1, xr.DataArray):
            if precision is None:
                # print(f"check xr {obj1} and {obj2}")
                message = f"{obj1}!={obj2}\n{obj1} and {obj2} do not match"
                assert xr.testing.assert_equal(obj1, obj2) is None
            else:
                # print(f"check xr {obj1} and {obj2} with precision {precision}")
                message = f"{obj1}!={obj2} with precision {precision}\n{obj1} and {obj2} do not match"
                assert xr.testing.assert_allclose(obj1, obj2, atol=10**(-precision)) is None
        elif isinstance(obj1, np.ndarray):
            if precision is None:
                # print(f"check np {obj1} and {obj2}")
                message = f"{obj1}!={obj2}\n{obj1} and {obj2} do not match"
                assert np.testing.assert_array_equal(obj1, obj2) is None
            else:
                # print(f"check np {obj1} and {obj2} with precision {precision}")
                message = f"{obj1}!={obj2} with precision {precision}\n{obj1} and {obj2} do not match"
                assert np.testing.assert_array_almost_equal(obj1, obj2, decimal=precision) is None
        else:
            if isinstance(obj1, (int, float)) and isinstance(obj2, (int, float)):
                if precision is None:
                    # print(f"check number equal {obj1} and {obj2}")
                    message = f"{obj1}!={obj2}\n{obj1} and {obj2} do not match"
                    assert np.testing.assert_equal(obj1, obj2) is None
                else:
                    # print(f"check number equal {obj1} and {obj2} with precision {precision}")
                    message = f"{obj1}!={obj2} with precision {precision}\n{obj1} and {obj2} do not match"
                    assert np.testing.assert_almost_equal(obj1, obj2, decimal=precision) is None
            else:
                # print(f"check equal {obj1} and {obj2}")
                message = f"{obj1}!={obj2}\n{obj1} and {obj2} do not match"
                assert obj1 == obj2
    except AssertionError:
        logging.info(message)
        return False
    return True
